#!/bin/bash

. tools/functions.sh

# Default variables
export VERBOSE=false
export OUTPUT="output"
export CACHE_FS="cache"
export ISO_NAME="nexus.iso"
export UBUNTU_DISTRIB="noble"
export ARCH="amd64"

helper() {
  echo "USAGE: $(basename "$0") envfile"
  echo "  envfile    path to an environment file containing the config to use when building the image"
  exit 1
}

export UBUNTU_REPO="https://archive.ubuntu.com/ubuntu"
export SQUASHFS_IMG="squash.rootfs"
export ROOTFS_DIR="$OUTPUT/rootfs"

SQUASHFS="$OUTPUT/squashfs"
ISO_DIR="$OUTPUT/iso"
BOOT_DIR="$ISO_DIR/boot"

TMP_KERNEL=$(mktemp /tmp/vmlinuz.XXXXXX)
TMP_INITRD=$(mktemp /tmp/initrd.XXXXXX)
TMP_SQUASHFS_IMG=$(mktemp /tmp/squashfs_img.XXXXXX)

if [ $# -ne 1 ]; then
  helper
fi

if [ -f "$1" ]; then
  source "$1" > /dev/null
  # prepend the CERT variable with the current directory (required by nexus-exam build process)
  export CERT=$(basename $(pwd))/$CERT
else
  echo "Error: file '$1' not found." >&2
  helper
fi

check_inside_container
check_environment_var
check_dependencies
CMDLINE="boot=nexus $RAM modprobe.blacklist=floppy"

echo "[Cleanup $ROOTFS_DIR]"
ROOTFS_PATH="${ROOTFS_DIR%/}"
child_mounts=$(findmnt -rn -o TARGET | grep -E "^${ROOTFS_PATH}/[^/]+$")

if [ -n "$child_mounts" ]; then
    echo "WARNING: The following mount points are directly under $ROOTFS_DIR:"
    echo "$child_mounts"
    echo
    read -p "Do you want to unmount these mount points? [y/N]: " confirmation
    case "$confirmation" in
        [yY][eE][sS]|[yY])
            while IFS= read -r mount_point; do
                echo "Unmounting $mount_point..."
                if ! umount "$mount_point"; then
                    echo "Error: failed to unmount $mount_point. Please unmount manually or reboot the system to avoid issues."
                    exit 1
                fi
            done <<< "$child_mounts"
            ;;
        *)
            echo "Cleanup aborted by user."
            exit 1
            ;;
    esac
fi
run_command rm -rf "$ROOTFS_DIR"

echo "[Cleanup $OUTPUT]"
run_command rm -rf "$OUTPUT"

echo "[Create filesystem...]"
mkdir -p "$ROOTFS_DIR"

if [ -z "$CACHE_FS" ]; then
  run_command debootstrap --arch="$ARCH" "$UBUNTU_DISTRIB" "$ROOTFS_DIR" $UBUNTU_REPO
  run_command mount -t proc /proc "$ROOTFS_DIR/proc"
  run_command mount --rbind /sys "$ROOTFS_DIR/sys"
  run_command mount --rbind /dev "$ROOTFS_DIR/dev"
  run_command mount --rbind /run "$ROOTFS_DIR/run"
  run_command mount -t tmpfs tmpfs "$ROOTFS_DIR/tmp"

  echo "[Installing packages...]"
  packages=$(tr '\n' ' ' < config/01-packages_install/packages)
  echo "deb $UBUNTU_REPO $UBUNTU_DISTRIB main universe" > "$ROOTFS_DIR/etc/apt/sources.list"
  run_command_chroot apt update
  # shim-signed required for UEFI secure boot
  run_command_chroot apt install -y --no-install-recommends "$packages"
else
  if [ ! -d "$CACHE_FS" ]; then
    mkdir "$CACHE_FS"
    run_command debootstrap --arch="$ARCH" "$UBUNTU_DISTRIB" "$CACHE_FS" $UBUNTU_REPO

    run_command mount -t proc /proc "$CACHE_FS/proc"
    run_command mount --rbind /sys "$CACHE_FS/sys"
    run_command mount --rbind /dev "$CACHE_FS/dev"
    run_command mount --rbind /run "$CACHE_FS/run"
    run_command mount -t tmpfs tmpfs "$CACHE_FS/tmp"

    echo "[Installing packages...]"
    packages=$(tr '\n' ' ' < config/01-packages_install/packages)
    echo "deb $UBUNTU_REPO $UBUNTU_DISTRIB main universe" > "$CACHE_FS/etc/apt/sources.list"
    run_command chroot "$CACHE_FS" apt update
    run_command chroot "$CACHE_FS" apt install -y --no-install-recommends "$packages"

    run_command umount -l "$CACHE_FS/proc"
    run_command umount -l "$CACHE_FS/sys"
    run_command umount -l "$CACHE_FS/dev"
    run_command umount -l "$CACHE_FS/run"
    run_command umount -l "$CACHE_FS/tmp"
  else
    echo "[Skip through cache]"
  fi

  run_command cp -r "$CACHE_FS"/. "$ROOTFS_DIR"/

  run_command mount -t proc /proc "$ROOTFS_DIR/proc"
  run_command mount --rbind /sys "$ROOTFS_DIR/sys"
  run_command mount --rbind /dev "$ROOTFS_DIR/dev"
  run_command mount --rbind /run "$ROOTFS_DIR/run"
  run_command mount -t tmpfs tmpfs "$ROOTFS_DIR/tmp"
fi

echo "[Copying configuration file...]"
run_command cp -rf config/02-customisation/* "$ROOTFS_DIR"

echo "[Checking for required dependencies...]"
if [ ! -d "/opt/limine" ]; then
  echo "Error: /opt/limine not found. Please install the Limine bootloader files in /opt/limine."
  exit 1
fi

if [ ! -f "$ROOTFS_DIR/usr/lib/shim/shimx64.efi.signed.latest" ] || [ ! -f "$ROOTFS_DIR/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed" ]; then
  echo "Error: required EFI files not found in $ROOTFS_DIR."
  exit 1
fi

if [ -z "$PXE_URL" ]; then
  EFIBOOT=$(mktemp /tmp/efiboot.img.XXXXXX)
  EFIBOOT_MOUNT=$(mktemp -d /tmp/efiboot.XXXXXX)
  echo "[EFI img preparation]"
  run_command mkdir -p "$ISO_DIR/boot/grub"
  run_command "sed 's/CMDLINE/$CMDLINE/g' config/00-bootloader/grub.cfg > $ISO_DIR/boot/grub/grub.cfg"
  SIZE_EFIBOOT=$(get_total_size "$ROOTFS_DIR/usr/lib/shim/shimx64.efi.signed.latest" "$ROOTFS_DIR/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed")
  OFFSET=$(( 150 * 1024 )) # Some space for fat header and grub.cfg file

  run_command fallocate -l $(( $SIZE_EFIBOOT + $OFFSET)) "$EFIBOOT"
  run_command mkfs.vfat "$EFIBOOT"
  run_command mount -o loop "$EFIBOOT" "$EFIBOOT_MOUNT"
  run_command mkdir -p "$EFIBOOT_MOUNT/EFI/BOOT"
  run_command cp config/00-bootloader/grub.efi.cfg "$EFIBOOT_MOUNT/EFI/BOOT/grub.cfg"
  run_command cp "$ROOTFS_DIR/usr/lib/shim/shimx64.efi.signed.latest" "$EFIBOOT_MOUNT/EFI/BOOT/BOOTx64.EFI"
  run_command cp "$ROOTFS_DIR/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed" "$EFIBOOT_MOUNT/EFI/BOOT/grubx64.efi"
  run_command umount "$EFIBOOT_MOUNT"
  run_command rmdir "$EFIBOOT_MOUNT"
fi

echo "[Post-install...]"
for script in config/03-post_install/*.sh; do
  echo "[Running $script...]"
  run_command "$script"
done

echo "[Moving kernel...]"
KERNEL_FILE=$(ls -1 "$ROOTFS_DIR/boot/vmlinuz-"* 2>/dev/null | head -n 1)
if [ -z "$KERNEL_FILE" ]; then
  echo "Error: no kernel file found in $ROOTFS_DIR/boot"
  exit 1
fi
run_command cp "$KERNEL_FILE" "$TMP_KERNEL"
run_command cp "$ROOTFS_DIR/boot/initrd.img" "$TMP_INITRD"
run_command rm -rf "$ROOTFS_DIR/boot"

echo "[Unmount subsystem...]"
run_command umount -l "$ROOTFS_DIR/proc"
run_command umount -l "$ROOTFS_DIR/sys"
run_command umount -l "$ROOTFS_DIR/dev"
run_command umount -l "$ROOTFS_DIR/run"
run_command umount -l "$ROOTFS_DIR/tmp"

echo "[Create squash.rootfs...]"
run_command mksquashfs "$ROOTFS_DIR" "$SQUASHFS" -comp zstd

if [ ! -z "$LUKS_PASSPHRASE" ]; then
  echo "[Encrypt squash.rootfs...]"
  SQUASHFS_SIZE=$(stat -c %s "$SQUASHFS")
  LUKS_HEADER_SIZE=$((1024 * 1024 * 16)) # 16 MiB header LUKS2
  TOTAL_SIZE=$((SQUASHFS_SIZE + LUKS_HEADER_SIZE))
  SECTOR_SIZE=512

  run_command fallocate -l "$TOTAL_SIZE" "$TMP_SQUASHFS_IMG"
  run_command bash -c "cat <<EOF | cryptsetup luksFormat '$TMP_SQUASHFS_IMG' --batch-mode
$LUKS_PASSPHRASE
EOF"

  run_command bash -c "cat <<EOF | cryptsetup luksOpen '$TMP_SQUASHFS_IMG' container
$LUKS_PASSPHRASE
EOF"
  run_command dd if="$SQUASHFS" of=/dev/mapper/container bs=4M
  rm -f "$SQUASHFS"
  run_command cryptsetup luksClose container

  if [ ! -z "$ADDITIONAL_KEYS" ]; then
    for key in $ADDITIONAL_KEYS; do
      key_file=$(mktemp)
      echo "$key" > "$key_file"
      run_command bash -c "cat <<EOF | cryptsetup luksAddKey '$TMP_SQUASHFS_IMG'
$LUKS_PASSPHRASE
$key
EOF"
      rm -f "$key_file"
    done
  fi
else
  mv "$SQUASHFS" "$TMP_SQUASHFS_IMG"
fi

if [ ! -z "$PXE_URL" ]; then
  run_command cp "$TMP_KERNEL" "$OUTPUT/vmlinuz"
  run_command cp "$TMP_INITRD" "$OUTPUT/initrd"
  run_command cp "$TMP_SQUASHFS_IMG" "$OUTPUT/$SQUASHFS_IMG"
  run_command cp "$ROOTFS_DIR/usr/lib/shim/shimx64.efi.signed.latest" "$OUTPUT/BOOTx64.EFI"
  run_command cp "$ROOTFS_DIR/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed" "$OUTPUT/grubx64.efi"
  CMDLINE="boot=pxe $RAM modprobe.blacklist=floppy"
  run_command "sed 's/CMDLINE/$CMDLINE/g' config/00-bootloader/grub.cfg > $OUTPUT/grub.cfg"
else
  echo "[Moving kernel && squashfs from rootfs to iso/boot...]"
  run_command mkdir -p "$BOOT_DIR"
  run_command cp "$TMP_KERNEL" "$BOOT_DIR/vmlinuz"
  run_command cp "$TMP_INITRD" "$BOOT_DIR/initrd"

  echo "[Create iso...]"
  run_command cp "$TMP_SQUASHFS_IMG" "$ISO_DIR/$SQUASHFS_IMG"
  run_command cp /opt/limine/*.bin "$ISO_DIR/"
  run_command cp /opt/limine/*.sys "$ISO_DIR/"
  run_command cp "$EFIBOOT" "$ISO_DIR/efiboot.img"
  run_command "sed 's/CMDLINE/$CMDLINE/g' config/00-bootloader/limine.conf > $ISO_DIR/boot/limine.conf"
  # IMPORTANT!!!
  # -volid value MUST MATCH the one in config/02-customisation/etc/initramfs-tools/scripts/nexus
  run_command xorriso -as mkisofs -R -r -J -b "limine-bios-cd.bin" \
        -no-emul-boot -boot-load-size 4 -boot-info-table -hfsplus \
        -apm-block-size 2048 --efi-boot "efiboot.img" \
        -efi-boot-part --efi-boot-image --protective-msdos-label \
        -volid "nexus-live-exam" \
        "$ISO_DIR" -o "$OUTPUT/$ISO_NAME"

  run_command limine bios-install "$OUTPUT/$ISO_NAME"
fi

chown -R 1000:1000 "$OUTPUT"

echo "[Cleanup temporary files...]"
rm -f "$TMP_KERNEL" "$TMP_INITRD" "$TMP_SQUASHFS_IMG" "$EFIBOOT"
